from ._impl import ImplInfo, get_impl_info, impl
from ._plugin import Plugin
from ._run import Run, active_run, log_asset, log_metrics
from ._spec import SpecInfo, spec
from .typed import MethodName, PluginId

__all__ = [
    "ImplInfo",
    "MethodName",
    "Plugin",
    "PluginId",
    "Run",
    "SpecInfo",
    "active_run",
    "get_impl_info",
    "impl",
    "log_asset",
    "log_metrics",
    "spec",
]



import functools
import inspect
from collections.abc import Callable, Mapping, Sequence
from typing import Any, Protocol, overload

import attrs

from liblaf import grapes

from .typed import MethodName


class Plugin(Protocol):
    def delegate(
        self,
        method: MethodName,
        args: Sequence[Any],
        kwargs: Mapping[str, Any],
        *,
        first_result: bool = False,
    ) -> Any: ...


@attrs.define
class SpecInfo:
    delegate: bool = attrs.field(default=True)
    first_result: bool = attrs.field(default=False)


@overload
def spec[C: Callable](
    func: C, /, *, delegate: bool = True, first_result: bool = False
) -> C: ...
@overload
def spec[C: Callable](
    *, delegate: bool = True, first_result: bool = False
) -> Callable[[C], C]: ...
def spec(
    func: Callable | None = None,
    /,
    *,
    delegate: bool = True,
    first_result: bool = False,
) -> Any:
    if func is None:
        return functools.partial(spec, delegate=delegate, first_result=first_result)

    info = SpecInfo(delegate=delegate, first_result=first_result)

    @grapes.decorator(attrs={"_self_spec": info})
    def wrapper(
        wrapped: Callable, instance: Plugin, args: tuple, kwargs: dict[str, Any]
    ) -> Any:
        if info.delegate:
            return instance.delegate(
                wrapped.__name__, args, kwargs, first_result=info.first_result
            )
        return wrapped(*args, **kwargs)

    return wrapper(func)


def collect_specs(cls: type[Plugin] | Plugin) -> dict[str, SpecInfo]:
    if isinstance(cls, type):
        cls = type(cls)
    return {
        name: method._self_spec  # noqa: SLF001
        for name, method in inspect.getmembers(
            cls, lambda m: getattr(m, "_self_spec", None) is not None
        )
    }



from typing import override

from liblaf.cherries import core

from ._playground import ProfilePlayground


class ProfileDefault(ProfilePlayground):
    @override  # impl Profile
    def init(self) -> core.Run:
        run: core.Run = super().init()
        return run



from . import config, core, meta, paths
from ._entrypoint import end, run, start
from .config import BaseConfig, input, output  # noqa: A004
from .core import Plugin, Run, active_run, log_asset, log_metrics
from .paths import (
    as_os_path,
    as_path,
    as_posix,
    data,
    entrypoint,
    exp_dir,
    git_root,
    git_root_safe,
    inputs,
    outputs,
    params,
    path,
    src,
)

__all__ = [
    "BaseConfig",
    "Plugin",
    "Run",
    "active_run",
    "as_os_path",
    "as_path",
    "as_posix",
    "config",
    "core",
    "data",
    "end",
    "entrypoint",
    "exp_dir",
    "git_root",
    "git_root_safe",
    "input",
    "inputs",
    "log_asset",
    "log_metrics",
    "meta",
    "output",
    "outputs",
    "params",
    "path",
    "paths",
    "run",
    "src",
    "start",
]



from typing import override

from liblaf import grapes
from liblaf.cherries import core


class Logging(core.Run):
    @override
    @core.impl
    def start(self, *args, **kwargs) -> None:
        profile = grapes.logging.profiles.ProfileCherries(
            handlers=[
                grapes.logging.rich_handler(),
                grapes.logging.file_handler(sink=self.plugin_root.exp_dir / "run.log"),
            ]
        )
        grapes.logging.init(profile=profile)

    @override
    @core.impl
    def end(self, *args, **kwargs) -> None:
        if (self.plugin_root.exp_dir / "run.log").exists():
            self.plugin_root.log_asset(self.plugin_root.exp_dir / "run.log")
        if (self.plugin_root.exp_dir / "run.log.jsonl").exists():
            self.plugin_root.log_asset(self.plugin_root.exp_dir / "run.log.jsonl")



import inspect
from collections.abc import Mapping, MutableMapping, Sequence
from typing import Any, Self

import attrs
import networkx as nx

from ._impl import ImplInfo, get_impl_info
from ._spec import SpecInfo
from .typed import MethodName


@attrs.define
class Plugin:
    plugins: dict[str, "Plugin"] = attrs.field(factory=dict)

    _plugin_parent: Self | None = attrs.field(default=None)
    _sort_plugins_cache: MutableMapping[MethodName, Sequence["Plugin"]] = attrs.field(
        factory=dict, init=False
    )

    @classmethod
    def plugin_id_cls(cls) -> str:
        return cls.__name__

    @property
    def plugin_id(self) -> str:
        return self.plugin_id_cls()

    @property
    def plugin_root(self) -> Self:
        if self._plugin_parent is None:
            return self
        return self._plugin_parent.plugin_root

    @property
    def specs(self) -> dict[str, SpecInfo]:
        return {
            name: method._self_spec  # noqa: SLF001
            for name, method in inspect.getmembers(
                type(self), lambda m: getattr(m, "_self_spec", None) is not None
            )
        }

    def delegate(
        self,
        method: MethodName,
        args: Sequence[Any] = (),
        kwargs: Mapping[str, Any] = {},
        *,
        first_result: bool = False,
    ) -> Any:
        results: list[Any] = []
        for plugin in self._sort_plugins_cache.get(method, []):
            result: Any = getattr(plugin, method)(*args, **kwargs)
            if result is None:
                continue
            if first_result:
                return result
            results.append(result)
        return results

    def register(self, plugin: "Plugin") -> None:
        plugin._plugin_parent = self  # noqa: SLF001
        self.plugins[plugin.plugin_id] = plugin

    def _prepare(self) -> None:
        for method in self.specs:
            self._sort_plugins_cache[method] = self._sort_plugins(
                method, refresh_cache=True
            )

    def _sort_plugins(
        self, method: str, *, refresh_cache: bool = False
    ) -> Sequence["Plugin"]:
        if refresh_cache or method not in self._sort_plugins_cache:
            plugin_infos: dict[str, ImplInfo] = {
                plugin_id: info
                for plugin_id, plugin in self.plugins.items()
                if (info := get_impl_info(getattr(plugin, method, None))) is not None
            }
            graph = nx.DiGraph()
            for plugin_id, impl_info in plugin_infos.items():
                graph.add_node(plugin_id)
                for after in impl_info.after:
                    graph.add_edge(after, plugin_id)
                for before in impl_info.before:
                    graph.add_edge(plugin_id, before)
            self._sort_plugins_cache[method] = tuple(
                plugin
                for plugin_id in nx.lexicographical_topological_sort(
                    graph, key=lambda node: plugin_infos[node].priority
                )
                if (plugin := self.plugins.get(plugin_id)) is not None
            )
        return self._sort_plugins_cache[method]



import enum
from collections.abc import Callable, Iterable
from pathlib import Path
from typing import Any

import pydantic

from liblaf import grapes
from liblaf.cherries import paths
from liblaf.cherries.typed import PathLike


class AssetKind(enum.StrEnum):
    INPUT = enum.auto()
    OUTPUT = enum.auto()


type PathProvider = (
    PathLike | Iterable[PathLike] | Callable[[PathLike], PathLike | Iterable[PathLike]]
)


class MetaAsset:
    kind: AssetKind
    _extra: PathProvider | None = None

    def __init__(self, kind: AssetKind, extra: PathProvider | None = None) -> None:
        self.kind = kind
        self._extra = extra

    def get_extra(self, value: Path) -> list[Path]:
        if self._extra is None:
            return []
        if callable(self._extra):
            extra: Iterable[PathLike] = grapes.as_iterable(self._extra(value))
            return [Path(p) for p in extra]
        extra: Iterable[PathLike] = grapes.as_iterable(self._extra)
        return [Path(p) for p in extra]


def get_assets(cfg: pydantic.BaseModel, kind: AssetKind) -> list[Path]:
    assets: list[Path] = []
    for name, info in type(cfg).model_fields.items():
        value: Any = getattr(cfg, name)
        if isinstance(value, pydantic.BaseModel):
            assets.extend(get_assets(value, kind))
        for meta in info.metadata:
            if isinstance(meta, MetaAsset) and meta.kind == kind:
                value: Path = Path(value)
                assets.append(value)
                assets.extend(meta.get_extra(value))
    return assets


def get_inputs(cfg: pydantic.BaseModel) -> list[Path]:
    return get_assets(cfg, AssetKind.INPUT)


def get_outputs(cfg: pydantic.BaseModel) -> list[Path]:
    return get_assets(cfg, AssetKind.OUTPUT)


def input(path: PathLike, extra: PathProvider | None = None, **kwargs) -> Path:  # noqa: A001
    field_info: pydantic.fields.FieldInfo = pydantic.Field(paths.data(path), **kwargs)  # pyright: ignore[reportAssignmentType]
    field_info.metadata.append(MetaAsset(kind=AssetKind.INPUT, extra=extra))
    return field_info  # pyright: ignore[reportReturnType]


def output(path: PathLike, extra: PathProvider | None = None, **kwargs) -> Path:
    field_info: pydantic.fields.FieldInfo = pydantic.Field(paths.data(path), **kwargs)  # pyright: ignore[reportAssignmentType]
    field_info.metadata.append(MetaAsset(kind=AssetKind.OUTPUT, extra=extra))
    return field_info  # pyright: ignore[reportReturnType]



from pathlib import Path

import attrs

from liblaf.cherries import paths

from ._plugin import Plugin
from ._spec import spec


@attrs.define
class Run(Plugin):
    """.

    References:
        1. [Experiment - Comet Docs](https://www.comet.com/docs/v2/api-and-sdk/python-sdk/reference/Experiment/)
        2. [Logger | ClearML](https://clear.ml/docs/latest/docs/references/sdk/logger)
        3. [MLflow Tracking APIs | MLflow](https://www.mlflow.org/docs/latest/ml/tracking/tracking-api/)
    """

    @property
    def exp_dir(self) -> Path:
        return paths.exp_dir(absolute=True)

    @spec
    def end(self, *args, **kwargs) -> None: ...

    @spec
    def log_asset(self, *args, **kwargs) -> None: ...

    @spec
    def log_metrics(self, *args, **kwargs) -> None: ...

    @spec(delegate=False)
    def start(self, *args, **kwargs) -> None:
        self._prepare()
        self.delegate("start", args, kwargs)


active_run: Run = Run()
log_asset = active_run.log_asset
log_metrics = active_run.log_metrics



import functools
from collections.abc import Callable, Iterable
from typing import Any, overload

import attrs

from liblaf import grapes

from .typed import PluginId


@attrs.define
class ImplInfo:
    after: Iterable[PluginId] = attrs.field(default=())
    before: Iterable[PluginId] = attrs.field(default=())
    priority: int = 0


@overload
def impl[C: Callable](
    func: C,
    /,
    *,
    priority: int = 0,
    after: Iterable[PluginId] = (),
    before: Iterable[PluginId] = (),
) -> C: ...
@overload
def impl[C: Callable](
    *,
    priority: int = 0,
    after: Iterable[PluginId] = (),
    before: Iterable[PluginId] = (),
) -> Callable[[C], C]: ...
def impl(
    func: Callable | None = None,
    /,
    priority: int = 0,
    after: Iterable[PluginId] = (),
    before: Iterable[PluginId] = (),
) -> Any:
    if func is None:
        return functools.partial(impl, priority=priority, after=after, before=before)
    info = ImplInfo(after=after, before=before, priority=priority)

    @grapes.decorator(attrs={"_self_impl": info})
    def wrapper(
        wrapped: Callable, _instance: Any, args: tuple, kwargs: dict[str, Any]
    ) -> Any:
        return wrapped(*args, **kwargs)

    return wrapper(func)


def get_impl_info(func: Callable | None) -> ImplInfo | None:
    if func is None:
        return None
    return getattr(func, "_self_impl", None)
